/*****************************************************************************

	volume-src-limited
		
    Copyright (C) 2002  Lucas Walter

  	lucasw@u.washington.edu
	http://students.washington.edu/lucasw

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

******************************************************************************/

#include <iostream>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "main.hpp"
#include "misc.hpp"
#include "objloader.hpp"
#include "camera.hpp"
#include "light.hpp"
#include "rigid.hpp"

using namespace std;

namespace app 
{
	string APP_NAME = "volume";
	
	string LICENSE = "\
This is free software, and you are welcome to redistribute\n\
it under the conditions specified by the GPL\n\n";
	
	//string LICENSE ="";
	
	float x,y,z;
	int roll,pitch,turn,relMouseX,relMouseY;
	bool rightButton;
	bool leftButton;
	
	bool play;	
	
	int frameTimeMs;
	int updateTimeMs;
	
	int frameOldCount;
	int frameNewCount; 
	
	int updateOldCount;
	int updateNewCount;
	
	camera *theCamera;
	
	#ifdef RIGID_VIEWING
	rigid *rigidTest;
	#endif
	
	light *light1;
	light *light2;
}

int main(int argc, char* argv[])
{
	
	using namespace app;
	
	cout << APP_NAME << "\
 (c) 2002-2003 Lucas Walter\n\
This software comes with ABSOLUTELY NO WARRANTY\n" \
<< LICENSE << "\
http://icculus.org/~lucasw\n\
lucasw@icculus.org\n\
	"; 
	
	cout << "\n\
Usage: \n" 
<< APP_NAME << " [FILE.obj] [OPTIONS]...\n\n\
Option flags\n\
	-step		Determines spacing of volume interior points\n\
	-coord		CARTESIAN, CYLINDRICAL, or SPHERICAL interior point orientation\n"
<<"	\n\n\n";	
	
	string loadFile;
	float moiStep = 0.2f;
	float Ks = 0.1f;
	float Kd = 0.2f;
	float Kt = 0.09;
	float forceDamp = 100.0f;
	float drag = 0.99f;
	objloader::coordSystem COORD_SYSTEM = objloader::CARTESIAN;
	
	loadFile = "obj/sphere2.obj";
	
	if (argc > 1)
		loadFile = argv[1];
	for (int i = 2; i < argc-1; i+=2) {
		string param = argv[i];
			   if ("-step"	== param) {
			moiStep = atof(argv[i+1]);		
		} else if ("-coord" == param)  {
			string coord = (argv[i+1]);
			if      (coord == "CARTESIAN") 
				COORD_SYSTEM = objloader::CARTESIAN;
			else if (coord == "SPHERICAL")
				COORD_SYSTEM = objloader::SPHERICAL;
			else if (coord == "CYLINDRICAL")
				COORD_SYSTEM = objloader::CYLINDRICAL;	
			else {
				cerr << "No such coordinate system " << coord << "\nUse CARTESIAN, CYLINDRICAL, or SPHERICAL\n";				
				return 1;
			}
		}
	}		
		
	theCamera = new camera("Main Viewer",camera::CENTERED);
	theCamera->move(0,0,0,1.00f,-1.00f,-10.30f);
	
	/// Main light source
	light1 = new light(theCamera,GL_LIGHT1); 
	light1->setName("light 1");
	light1->move(0,0,0,40.0f,-10.0f,5.0f);

	light1->Kd[0] = 0.5f;
	light1->Kd[1] = 0.3f;
	light1->Kd[2] = 0.1f;
	
	light2 = new light(theCamera,GL_LIGHT2);
	light2->setName("light 1");
	light2->move(0,0,0,0.50f,5.0f,0.2f);
	
	light2->Kd[0] = 0.8f;
	light2->Kd[1] = 0.8f;
	light2->Kd[2] = 0.8f;
	
	vector<light*> lights;
	lights.push_back(light1);
	lights.push_back(light2);
	
	#ifdef RIGID_VIEWING
	try { rigidTest = new rigid( loadFile, moiStep, COORD_SYSTEM); }
	catch (fileNotFound) { return 1; }
	#endif
	
	play = true;
		
	x = y = z = 0;	
	
	/// Screen dimensions
	int width = 800;
	int height = 600;
	initSDL(width,height);
	initGL(width,height);

	updateTimeMs = (1000/25);  /// 25 screen and control updates per second

	frameTimeMs = (1000/10);  ///  10 physics/springs/etc. updates a second
	
	updateOldCount = 0;
	frameOldCount = 0;
	
	cout << "frameTimeMs " << frameTimeMs << "\n"; //>
	cout << "updateTimeMs " << updateTimeMs << "\n"; //>
	
	mainLoop();

	return 0;
}

void initGL(int width, int height)	       
{	
	//glEnable(GL_TEXTURE_2D);              
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	//glClearDepth(1.0);				         
	glDepthFunc(GL_LEQUAL);			    
	glEnable(GL_DEPTH_TEST);	
	glShadeModel(GL_SMOOTH);		

	#ifdef FOG
	vector3f fogColor(0.1f,0.1f,0.3f);
	glEnable(GL_FOG);
	glFogfv(GL_FOG_COLOR, fogColor.vertex);
	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, 8.0f);
	glFogf(GL_FOG_END, 12.0f);
	#endif
		
	glDisable(GL_LIGHTING);
	//glEnable(GL_COLOR_MATERIAL);
	glLineWidth(1.0f);	
	glEnable (GL_LINE_SMOOTH);
	glCullFace( GL_FRONT );
	glEnable (GL_CULL_FACE); 
	
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	
	glViewport(0, 0, width, height);
   	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();				

	gluPerspective(45.0f,(GLfloat)width/(GLfloat)height,1.0f,1000.0f);	

	glMatrixMode(GL_MODELVIEW);
}

void initSDL(int width, int height)
{	
	srand(time(NULL));

	if ( SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER ) < 0 ) {
		cerr << "Unable to initialize SDL: " << SDL_GetError() << "\n"; //>
      		exit(1);
	}

	atexit(SDL_Quit);

	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );

	unsigned int flags = SDL_OPENGL;// |SDL_RESIZABLE; // |SDL_RESIZABLE
	// flags |= SDL_FULLSCREEN;

   	if ( SDL_SetVideoMode(width, height, 32, flags) == NULL ) {
		cerr << "Unable to create OpenGL screen: " << SDL_GetError() << "\n"; 
		SDL_Quit();
		exit(2);
	}
	
	SDL_WM_SetCaption(app::APP_NAME.c_str(), NULL);

	//SDL_WM_GrabInput(SDL_GRAB_ON);
	//SDL_ShowCursor(SDL_DISABLE);
}

void update(Uint32 ms)					
{
	using namespace app;
	
	frameNewCount = updateNewCount = SDL_GetTicks();
	
	if ( (play) && ((frameNewCount - frameOldCount) > frameTimeMs) ) {
		//cout << "(frameNewCount - frameOldCount)  " << (frameNewCount - frameOldCount)  <<"\n";
		//cout << "frameTimeMs "<< frameTimeMs<< "\n";
		
		frameOldCount = frameNewCount;
	}
	
	if ( (updateNewCount - updateOldCount) > updateTimeMs ) {
		//cout << "(updateNewCount - updateOldCount) " <<(updateNewCount - updateOldCount) <<"\n";
		//cout << "updateTimeMs "<< updateTimeMs << "\n";
		theCamera->update();
		
		if (play) {
			#ifdef RIGID_VIEWING		
			rigidTest->update(updateTimeMs);
			#endif
		}
		
		SDL_GetRelativeMouseState(&relMouseX,&relMouseY);
		
		pitch = 0;
		turn = 0;
		roll = 0;
	
		if (leftButton) {
			pitch = relMouseY;
			turn  = relMouseX;
		} else if (rightButton) {
			roll = relMouseX;
		}
			
		theCamera->move(pitch,turn,roll,x,y,z);
		
		draw();
		
		updateOldCount = updateNewCount;
	}	
}

void draw()
{	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();
	
	using namespace app;
	
	theCamera->look();

	light1->draw();
	light2->draw();

	#ifdef RIGID_VIEWING
	rigidTest->draw();
	#endif
	
	glFlush();

	SDL_GL_SwapBuffers();
	
	glPopMatrix();

}

void handleInput(SDL_Event *E)
{
	using namespace app;
	
	rightButton = false;
	float move_inc = 0.4f;
	//int deg_inc = 6;	

	switch(E->type) {
		case SDL_MOUSEBUTTONDOWN:

			switch( E->button.button) {
				case SDL_BUTTON_LEFT:
					leftButton= true;

						break;
						
				case SDL_BUTTON_RIGHT:
					rightButton= true;

						break;
					default:
						break;
			}
			break;
					
		case SDL_MOUSEBUTTONUP:
			switch(E->button.button)
			{
				case SDL_BUTTON_LEFT:
					leftButton=false;
		
					break;
				case SDL_BUTTON_RIGHT:
					rightButton=false;
		
					break;
				default:
					break;
			}
			break;
					
	case SDL_KEYDOWN:
		switch( E->key.keysym.sym ) {
			
		case SDLK_ESCAPE:
			terminateApp();
			break;

		case SDLK_F1:			
			//ToggleFullscreen();	 
			break;
		case SDLK_F2:			
		 
			break;
			
		case SDLK_KP8:
			y += move_inc;
			break;
		case SDLK_KP2:
			y -= move_inc;
			break;
			
		case SDLK_KP5:
			z += move_inc;
			break;
		case SDLK_KP0:
			z -= move_inc;
			break;
					
		case SDLK_KP6:
			x += move_inc;
			break;
		case SDLK_KP4:
			x -= move_inc;
			break;
			
		case SDLK_SPACE:
			play = !play;
			break;
		
		case SDLK_UP:

		break;
				
		case SDLK_RIGHT:

		break;
		
		case SDLK_LEFT:	

		break;

		default:
			break;
			
		}
		break;
		
	case SDL_KEYUP:	
			
		switch( E->key.keysym.sym ){
				
		case SDLK_KP8:
			y -= move_inc;
			break;
		case SDLK_KP2:
			y += move_inc;
			break;
						
		case SDLK_KP5:
			z -= move_inc;
			break;
		case SDLK_KP0:
			z += move_inc;
			break;
					
		case SDLK_KP6:
			x -= move_inc;
			break;
		case SDLK_KP4:
			x += move_inc;
			break;

		case SDLK_TAB:
			
			break;
				
		case SDLK_RIGHT:
glCullFace( GL_BACK );
		break;
		
		case SDLK_LEFT:
	
		break;
		
		case SDLK_UP:

		break;
		
		case SDLK_DOWN:
	
		break;
		
		case SDLK_s:
			#ifdef RIGID_VIEWING
			rigidTest->drawSurface = !rigidTest->drawSurface;
		
			//glCullFace( GL_BACK );
			//glEnable( GL_CULL_FACE );
			#endif			
		break;
		
		case SDLK_p:
			#ifdef RIGID_VIEWING
			if (rigidTest->drawInteriorPoints) {
				rigidTest->drawInteriorPoints = false;
				rigidTest->drawInteriorSpheres = true;
			} else if (rigidTest->drawInteriorSpheres) {
				rigidTest->drawInteriorPoints = false;
				rigidTest->drawInteriorSpheres = false;
			} else {
				rigidTest->drawInteriorPoints = true;
				rigidTest->drawInteriorSpheres = false;
			}
			#endif			
		break;
			
		default:
			break;				
		}
			
		default:
			break;			
	}

	return;
}

void terminateApp(void)                                                 
{
        static SDL_Event Q;

        Q.type = SDL_QUIT;

        if(SDL_PushEvent(&Q) == -1) {
                cerr << "SDL_QUIT event can't be pushed: " << SDL_GetError() << "\n";         
                exit(1);
        }

        return;
 
}

int mainLoop()
{
	bool isProgramLooping = true;			
	SDL_Event E;

	while(isProgramLooping)
	{
		if(SDL_PollEvent(&E)) {
			switch(E.type) {
				
			case SDL_QUIT:
				isProgramLooping = false;
				break;				
			

			case SDL_VIDEORESIZE:
			//	reshapeGL(E.resize.w, E.resize.h);
				break;


			case SDL_ACTIVEEVENT:		
/*				if(E.active.state & SDL_APPACTIVE)
					{
						if(E.active.gain)
						{
							AppStatus.Visible = true;
						}
						else {
							AppStatus.Visible = false;				
						}
					}
					
					if(E.active.state & SDL_APPMOUSEFOCUS) {
						if(E.active.gain) {
							AppStatus.MouseFocus = true;	
						} else {
							AppStatus.MouseFocus = false;
						}
					
					if(E.active.state & SDL_APPINPUTFOCUS) {
						if(E.active.gain) {
							AppStatus.KeyboardFocus = true;	
						} else { 
							AppStatus.KeyboardFocus = false;
						}
*/	
					break;						
				
			case SDL_MOUSEBUTTONDOWN:
			case SDL_MOUSEBUTTONUP:
			case SDL_KEYDOWN:					
			case SDL_KEYUP:					
					handleInput(&E);
					break;	
			}
		} else {
			
			update(0);

		}
		
	}
			
	return 0;		
}
